/*
  Ncr4x20Vfd Library - Implementation

  Originally created 6 Feb 2011

  This file is in the public domain.  
*/

#include "Ncr4X20Vfd.h"
#include "WProgram.h"

//static
unsigned long Ncr4X20Vfd::_startupTime = 0;

Ncr4X20Vfd::Ncr4X20Vfd(uint8_t strobePin, uint8_t busyPin,
                       uint8_t data0Pin, uint8_t data1Pin, uint8_t data2Pin, uint8_t data3Pin,
                       uint8_t data4Pin, uint8_t data5Pin, uint8_t data6Pin, uint8_t data7Pin)
{
  if (_startupTime == 0)
  {
	// We need to init the startup timeout. The display takes a little while to stabilize at startup.
	reInitStartupTimeouts();
  }
	
  _flagBeginHasBeenCalled = false;
  _flagLastWaitForNotBusyFailed = false;
  _flagMightBeInReset = true; // They display may have just been powered on.  
  _flagHadTimeoutError = false;
  
  _strobePin = strobePin;
  _busyPin = busyPin;
  
  _dataPins[0] = data0Pin;
  _dataPins[1] = data1Pin;
  _dataPins[2] = data2Pin;
  _dataPins[3] = data3Pin;
  _dataPins[4] = data4Pin;
  _dataPins[5] = data5Pin;
  _dataPins[6] = data6Pin;
  _dataPins[7] = data7Pin;

  // Initalize pin settings.
  digitalWrite(_strobePin, HIGH);
  pinMode(_strobePin, OUTPUT);

  if (_busyPin != DO_NOT_USE_PIN)
  {
    pinMode(_busyPin, INPUT);
  }
  
  // Init all data pins to input. The data lines
  // on the display are bidirectional so we only want to drive them
  // when needed. This prevents both devices driving the lines
  // with high currents for extended periods in cases where an error
  // is made.
  for (int i=0; i<8; i++)
  {
    uint8_t pin = _dataPins[i];
    if (pin != DO_NOT_USE_PIN)
    {
      pinMode(pin, INPUT);  
    }
  }
  
  _lastCharacter = 0;
}


/* 
  Private Methods
*/
  
void Ncr4X20Vfd::enterCommand()
{
  // Do nothing here. If we were really keen we could
  // handle the case where we need to emit characters
  // in the middle of a user escape sequence but it
  // is much simpler just to say that we don't support
  // that.
}

void Ncr4X20Vfd::leaveCommand()
{
  // Do nothing here. See comment in enterCommand.
}

// Sends a simple one byte command code with proper enter/leave.
void Ncr4X20Vfd::handleSimpleOneCharCommand(uint8_t character)
{
  enterCommand();
  write(character);
  leaveCommand();
}

void Ncr4X20Vfd::handleSimpleTwoCharCommand(uint8_t char0, uint8_t char1)
{
  enterCommand();
  write(char0);
  write(char1);  
  leaveCommand();
}
  
bool Ncr4X20Vfd::waitForNotBusy()
{
  bool result = true;
	
  if (_startupTime > 1)
  {
	  // Timeout is valid, wait for it to complete.
	  unsigned long waitTime = (_busyPin == DO_NOT_USE_PIN) ? STARTUP_TIME_IN_MILLISECONDS : HARD_POWERUP_TIME_IN_MILLISECONDS;
	  while ((millis() - _startupTime) < waitTime)
	  {
	  }
	  // 1 means we have already waited for timeout to complete.
	  // We won't wait again unless reInitStartupTimeouts() is called.
	  _startupTime = 1; 	  
  }
  
  if (_busyPin != DO_NOT_USE_PIN) // Never wait of we don't have a pin to read.
  {
    unsigned long startMicros = micros();
    unsigned long startMillis = millis();

    for (;;)
    {
      if (digitalRead(_busyPin) == LOW)
      {
        // Display is no longer busy.
        break;
      }
      
      if (_flagMightBeInReset)
      {
        // Much longer timeout when resetting.
        if ( (millis()-startMillis) > STARTUP_TIME_IN_MILLISECONDS)
        {  
          result = false;
          break;
        }
      }
      else if (_flagLastWaitForNotBusyFailed || // Don't wait a second time
               ((micros()-startMicros) > WAIT_FOR_NOT_BUSY_TIMEOUT_IN_MICROSECONDS) )
      {
        result = false;
        break;
      }
    }    
  }
  
  if (result)
  {
    _flagLastWaitForNotBusyFailed = false;
  }
  else
  {
    // Save off so we don't wait next time.
    _flagLastWaitForNotBusyFailed = true;
    // Store this off so our unit tests can see if there was any errors.
    _flagHadTimeoutError = true;
  }

  // Only wait once for a reset.
  _flagMightBeInReset = false;

  return result;
}  

/*
  Public Methods
*/

void Ncr4X20Vfd::begin()
{
  _flagBeginHasBeenCalled = true;

  _flagMightBeInReset = true; // They display may have just been powered on.  
  
  // Set display to default state so app runs correctly after a
  // reload of the sketch.
  write(ESC_CHARACTER);
  write(DISPLAY_ON_SECOND_CHARACTER);
  write(HOME_CURSOR_CHARACTER);
  write(CLEAR_SCREEN_CHARACTER);
  write(CHARACTER_SET_INTERNATIONAL_CHARACTER);
  write(VERTICAL_SCROLL_OFF_CHARACTER);
}
  
void Ncr4X20Vfd::clear()
{
  handleSimpleTwoCharCommand(CLEAR_SCREEN_CHARACTER, HOME_CURSOR_CHARACTER);
}

void Ncr4X20Vfd::home()
{
  handleSimpleOneCharCommand(HOME_CURSOR_CHARACTER);
}
  
void Ncr4X20Vfd::noDisplay()
{
  handleSimpleTwoCharCommand(ESC_CHARACTER, DISPLAY_OFF_SECOND_CHARACTER);
}

void Ncr4X20Vfd::display()
{
  handleSimpleTwoCharCommand(ESC_CHARACTER, DISPLAY_ON_SECOND_CHARACTER);
}

void Ncr4X20Vfd::noBlink()
{
  handleSimpleOneCharCommand(CURSOR_OFF_CHARACTER);
}

void Ncr4X20Vfd::blink()
{
  handleSimpleOneCharCommand(CURSOR_ON_CHARACTER);
}

void Ncr4X20Vfd::noCursor()
{
  // Identical to noBlink.
  handleSimpleOneCharCommand(CURSOR_OFF_CHARACTER);
}

void Ncr4X20Vfd::cursor()
{
  // Identical to blink.
  handleSimpleOneCharCommand(CURSOR_ON_CHARACTER);
}
  
void Ncr4X20Vfd::setCursor(uint8_t column, uint8_t row)
{
  enterCommand();
  
  // Move to top left corner of display.
  write(HOME_CURSOR_CHARACTER);

  // Move to row.
  for (uint8_t i=0; i<row; i++)
  {
    write(LINE_FEED_CHARACTER);
  }

  // Move to column.
  for (uint8_t i=0; i<column; i++)
  {
    write(CURSOR_TAB_CHARACTER);
  }
  
  leaveCommand();
}
  
//virtual
void Ncr4X20Vfd::write(uint8_t character)
{
  if (_dataPins[7] == DO_NOT_USE_PIN)
  {
    // See if these will collide with the control code range.
    if (character>=128 && character<(128 + ' '))
    {
      // Force it to a space so there aren't unintended effects
      // from sending unexpected control codes.
      character = ' ';
    }
  }
	
  // Don't do anything until begin has been called. This is to enforce calling
  // of the begin() method in case we need to put something critical in there
  // later.
  if (_flagBeginHasBeenCalled)
  {
    if (waitForNotBusy())
    { 
      uint8_t tmpCharacter = character;
      for (int i=0; i<8; i++)
      {
        uint8_t pin = _dataPins[i];
        if (pin != DO_NOT_USE_PIN)
        {
          // Set pin value.
          digitalWrite(pin, (tmpCharacter & 1) ? HIGH : LOW);
          tmpCharacter >>= 1;
          
          // Enable pin output.
          pinMode(pin, OUTPUT);  
        }
      }
      
      delayMicroseconds(DATA_SETUP_TIME_IN_MICROSECONDS);

      // Write data to display. Strobe is active low.      
      digitalWrite(_strobePin, LOW);
      
      delayMicroseconds(STROBE_HOLD_TIME_IN_MICROSECONDS);
      
      // Restore strobe line.      
      digitalWrite(_strobePin, HIGH);
      
      for (int i=0; i<8; i++)
      {
        uint8_t pin = _dataPins[i];
        if (pin != DO_NOT_USE_PIN)
        {
          // Disable pin output.
          pinMode(pin, INPUT);  
        }
      }
            
      if (_busyPin == DO_NOT_USE_PIN)
      {
        delayMicroseconds(NO_BUSY_PIN_WAIT_TIME_IN_MICROSECONDS);
      }
    }
    
    // Update last character.
    if (_lastCharacter == ESC_CHARACTER)
    {
      if (character == DISPLAY_DIAGNOSTIC_SECOND_CHARACTER)
      {
        // We have just reset the display.
        _flagMightBeInReset = true;
      }
    }
    _lastCharacter = character;
  }
}

void Ncr4X20Vfd::setDefaultCharacterSet()
{
  handleSimpleOneCharCommand(CHARACTER_SET_INTERNATIONAL_CHARACTER);
}

void Ncr4X20Vfd::setJapaneseCharacterSet()
{
  handleSimpleOneCharCommand(CHARACTER_SET_JAPANESE_CHARACTER);
}

void Ncr4X20Vfd::setCodePage850CharacterSet()
{
  handleSimpleOneCharCommand(CHARACTER_SET_CODE_PAGE_850_CHARACTER);
}

void Ncr4X20Vfd::noVerticalScroll()
{
  handleSimpleOneCharCommand(VERTICAL_SCROLL_OFF_CHARACTER);
}

void Ncr4X20Vfd::verticalScroll()
{
  handleSimpleOneCharCommand(VERTICAL_SCROLL_ON_CHARACTER);
}

void Ncr4X20Vfd::startDiagnosticSequence()
{
  handleSimpleTwoCharCommand(ESC_CHARACTER, DISPLAY_DIAGNOSTIC_SECOND_CHARACTER);
}

void Ncr4X20Vfd::endDiagnosticSequence()
{
  // The "enable display" command brings the display out of the diagnostic sequence.
  display();
}

bool Ncr4X20Vfd::getError()
{
  bool result = _flagHadTimeoutError;
  _flagHadTimeoutError = false;
  return result;
}
			 
void Ncr4X20Vfd::reInitStartupTimeouts()
{
  _startupTime = millis();
  if (_startupTime <= 1) // Values 0 and 1 have special meaning.
  {
    _startupTime = 2; // 2 will be close enough for our timeouts.
  }
}
