/*
  Ncr4x20Vfd Library - Unit Tests
  
  Tests funationality of library for 
  the NCR 7454 / Futaba M204SD01B
  4x20 VFD module.
  
  This sketch is controlled via the serial
  monitor. An ATmega328 or better microcontroller
  is required. The ATmega168 and ATmega8 do
  not have enough SRAM and will crash trying 
  to run this sketch.
  
  Some tests require the display power
  to be switched via pin 13. This can
  be done with a high-side p-channel
  mosfet for example.
  
  TODO: Insert link to wiring tutoral.
  TODO: Add link to Arduino reference page.
  
  Originally created 9 Feb 2011

  This file is in the public domain.
*/

#include <Ncr4X20Vfd.h>
#include <stdlib.h> // for malloc and free
#include <avr/pgmspace.h> // Allows storing of strings in program memory which is good, we have a lot of them.

// See Ncr4X20VfdTutorialPins.h for other board defines.
#define NCR_4X20_VFD_ARDUINO_PINOUT
#include "Ncr4X20TutorialPins.h"

#define USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY 0

static const uint8_t DISPLAY_POWER_SWITCH_PIN = 13;

static const int DISPLAY_CHARS_X = 20;
static const int DISPLAY_CHARS_Y = 4;
static const int DISPLAY_CHARS_TOTAL = DISPLAY_CHARS_X * DISPLAY_CHARS_Y;

typedef bool (*TestFunction)(bool, bool);

struct UnitTest {

  const prog_char * _title;
  const prog_char * _instructions;
  TestFunction _testFunction;
};

static prog_char basicTextTestTitle [] PROGMEM = 
    "Basic Text Test";
static prog_char basicTextTestInstructions [] PROGMEM = 
    "This test will fill the screen with the following text:\n"
    "\n"
    "*** NCR 4x20 VFD ***\n"
    "* Futaba M204SD01B *\n"
    "********************\n"
    "titaniumstudios.com \n";
    
static prog_char clearTestTitle [] PROGMEM = 
    "clear() Test";
static prog_char clearTestInstructions [] PROGMEM = 
    "This test will clear the screen.\n";

static prog_char homeTestTitle [] PROGMEM = 
    "home() Test";
static prog_char homeTestInstructions [] PROGMEM = 
    "This test will print text, home the cursor, then print more text,\n" 
    "overwriting part of the original. If the test works correctly this\n"
    "will be displayed:\n"
    "\n"
    "YOU SHOULD SEE\n"
    "THIS\n"
    "AND\n"
    "THIS\n"
    "\n"
    "If the test does not work correctly, this may be displayed:\n"
    "\n"
    "YOU SHOULD\n"
    "NOT\n"
    "SEE\n"
    "THIS\n";
    
static prog_char noDisplayTestTitle [] PROGMEM = 
    "noDisplay() Test";
static prog_char noDisplayTestInstructions [] PROGMEM = 
    "You should see \"Waiting 2 seconds.\" followed by\n"
    "2 seconds of blank screen, followed by \"And we are back.\"\n"
    "in the top left of the screen.\n";
    
static prog_char blinkCursorTestTitle [] PROGMEM = 
    "blink()/cursor() Test";
static prog_char blinkCursorTestInstructions [] PROGMEM = 
    "You should see numbers count slowly and fill the display.\n"
    "On the first and third rows the cursor should blink.\n";
    
static prog_char setCursorTestTitle [] PROGMEM = 
    "setCursor() Test";
static prog_char setCursorTestInstructions [] PROGMEM = 
    "The screen should full with sequential characters in random fill order,\n"
    "stay filled for 4 seconds, then fill with spaces in random order.\n";
    
static prog_char characterSetsTestTitle [] PROGMEM = 
    "Character Sets Test";
static prog_char characterSetsTestInstructions [] PROGMEM = 
    "Slowly prints out all 3 character sets, each in a different row.\n";
  
static prog_char verticalScrollTestTitle [] PROGMEM = 
    "verticalScroll() Test";
static prog_char verticalScrollTestInstructions [] PROGMEM = 
    "Prints 12 rows that scroll, then 12 rows that wrap, 2x over.\n";
  
static prog_char startDiagnosticSequenceTestTitle [] PROGMEM = 
    "startDiagnosticSequence() Test";
static prog_char startDiagnosticSequenceTestInstructions [] PROGMEM = 
    "Starts the diagnostic sequence and runs it for 15 seconds.\n"
    "After completion, display should read \"Ended diagnostic sequence.\"\n";

static prog_char highBitCharacterTestTitle [] PROGMEM = 
    "data7 Pin Character Test";
static prog_char highBitCharacterTestInstructions [] PROGMEM = 
    "The first two lines of the display will contain A-Z\n"
    "and 0-9. When not using the data7 pin, the bottom two\n"
    "lines will contain the same characters. When using the\n"
    "data7 pin, the bottom two lines will contain extended\n"
    "characters.\n";

static prog_char loadTestSequenceCharactersTestTitle [] PROGMEM = 
    "Load Test Using Many Characters";
static prog_char loadTestSequenceCharactersTestInstructions [] PROGMEM = 
    "Prints each character 80x, ending in a y character with two dots.\n"
    "When not using the data7 pin, then end character will by a Y with\n"
    "two horizontal lines.\n";

static prog_char loadTestWithClearsTestTitle [] PROGMEM = 
    "Load Test Using Clears";
static prog_char loadTestWithClearsTestInstructions [] PROGMEM = 
    "Prints 800 digits with clears after each,\n"
    "then prints A-T across the top line.\n"
    "\n"
    "The display may be inaccurate during the test but should\n"
    "be correct at the end.\n";

static prog_char loadTestWithScrollingTestTitle [] PROGMEM = 
    "Load Test Using Scrolling";
static prog_char loadTestWithScrollingTestInstructions [] PROGMEM = 
    "Prints 800 digits with a scroll after each,\n"
    "then prints A-T across the bottom line.\n"
    "\n"
    "The display may be inaccurate during the test but should\n"
    "be correct at the end. At the end, digits 8 and 9 should\n"
    "form a diagonal in the top right corner.\n";

static prog_char hardResetTestTitle [] PROGMEM = 
    "Hard Power Cycle Test";
static prog_char hardResetTestInstructions [] PROGMEM = 
#if USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY
    "This test requires a high side transistor switch for display\n"
    "power to be connected to (switched by) pin 13. Skip this test\n"
    "unless you are 100% sure how to do this as incorrect\n"
    "connections may damage something.\n"
    "\n"
#endif
    "Tests that the display properly waits and functions after\n"
    "hard-powering up the display. After power up, \"YOU SHOULD\n"
    "SEE THIS.\" should be displayed.\n";
    
static prog_char forceTimeoutTestTitle [] PROGMEM = 
    "Force Timeout Test";
static prog_char forceTimeoutTestInstructions [] PROGMEM = 
#if USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY
    "This test requires a high side transistor switch for display\n"
    "power to be connected to (switched by) pin 13. Skip this test\n"
    "unless you are 100% sure how to do this as incorrect\n"
    "connections may damage something.\n"
    "\n"
#endif
    "Tests that the timeouts actually timeout correctly by power cycling\n"
    "the display behind the library's back. When complete, \"YOU SHOULD\n"
    "SEE THIS.\" should be displayed.\n";
 
// The Arduino envrionment doesn't supply new and delete so we must implement it.
void* operator new(size_t size)
{ 
  return malloc(size); 
}

void operator delete(void* ptr) 
{
  if (ptr)
  {
    free(ptr);
  }
} 

char getCharacter()
{
  for (;;)
  {
    if (Serial.available()) 
    {
      return Serial.read();
    }
  }
  
  return 0;
}

void getString(char *destBuffer, int destBufferLength)
{
  if (destBufferLength == 0)
  {
    destBuffer[0] = 0;
  }
  else if (destBufferLength > 1)
  {
    destBuffer[0] = getCharacter();

    int numChars = 1;
    unsigned long lastCharacterTime = millis();
  
    // Collect characters until we get a 100ms gap or buffer fills up.
    while (numChars<(destBufferLength-1) && (millis()-lastCharacterTime)<100)
    {
      if (Serial.available()) 
      {
        destBuffer[numChars] = Serial.read();
        lastCharacterTime = millis();
        numChars++;
      }
    }
    
    destBuffer[numChars] = 0;
  }
}

void serialPrintFromProgmem(const prog_char string[])
{
  if (string)
  {
    char character;
    while((character = pgm_read_byte(string++)))
    {
      Serial.write(character);
    }
  }
}

Ncr4X20Vfd *createVfd(bool useBusyPin, bool useData7Pin)
{
  if (!useData7Pin)
  {
    // Set pin low as if it were tied to ground.
    digitalWrite(NCR_4X20_VFD_DATA7_PIN, LOW);
    pinMode(NCR_4X20_VFD_DATA7_PIN, OUTPUT);  
  }
  
  return new Ncr4X20Vfd(
    NCR_4X20_VFD_STROBE_PIN,
    useBusyPin ? NCR_4X20_VFD_BUSY_PIN : Ncr4X20Vfd::DO_NOT_USE_PIN,
    NCR_4X20_VFD_DATA0_PIN,
    NCR_4X20_VFD_DATA1_PIN,
    NCR_4X20_VFD_DATA2_PIN,
    NCR_4X20_VFD_DATA3_PIN,
    NCR_4X20_VFD_DATA4_PIN,
    NCR_4X20_VFD_DATA5_PIN,
    NCR_4X20_VFD_DATA6_PIN,
    useData7Pin ? NCR_4X20_VFD_DATA7_PIN : Ncr4X20Vfd::DO_NOT_USE_PIN);
}

void fillBufferWithRandomOrderIndices(uint8_t *indices, int numIndices)
{
  // Fill with sequential numbers.
  for (int i=0; i<numIndices; i++)
  {
    indices[i] = i;
  }
  
  // Randomize by performing random swaps.
  for (int i=0; i<numIndices; i++)
  {
    int otherIndex = random(numIndices);
    uint8_t otherValue = indices[otherIndex];
    indices[otherIndex] = indices[i];
    indices[i] = otherValue;
  }
}

void clearAndHomeDisplay(Ncr4X20Vfd *vfd)
{
  vfd->clear();
  vfd->home();
}

bool basicTextTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  vfd->print("*** NCR 4x20 VFD ***"
             "* Futaba M204SD01B *"
             "********************"
             "titaniumstudios.com ");
               
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool clearTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  vfd->clear();
  
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool homeTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  vfd->print("YOU SHOULD\r\n"
             "NOT\r\n"
             "SEE\r\n"
             "THIS");

  vfd->home();
             
  vfd->print("YOU SHOULD SEE\r\n"
             "THIS\r\n"
             "AND");
  
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool noDisplayTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  vfd->print("Waiting 2 seconds.\r\n");
  
  delay(2000);
  
  vfd->noDisplay();
  
  vfd->print("You should not see\r\n"
             "this.\r\n");
  
  delay(2000);

  vfd->display();

  clearAndHomeDisplay(vfd);
  
  vfd->print("And we are back.");
    
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

void blinkCursorTestPrintNumbers(Ncr4X20Vfd *vfd)
{
  for (int i=0; i<20; i++)
  {
    delay(250);
    vfd->print(i%10);
  } 
}

bool blinkCursorTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  vfd->cursor();
  blinkCursorTestPrintNumbers(vfd);

  vfd->noCursor();
  blinkCursorTestPrintNumbers(vfd);
  
  vfd->blink();
  blinkCursorTestPrintNumbers(vfd);
  
  vfd->noBlink();
  blinkCursorTestPrintNumbers(vfd);
    
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

void setCursorTestFill(Ncr4X20Vfd *vfd, bool useSpaces)
{
  uint8_t indices[DISPLAY_CHARS_TOTAL];
  
  fillBufferWithRandomOrderIndices(indices, DISPLAY_CHARS_TOTAL);
  
  for (int i = 0; i < DISPLAY_CHARS_TOTAL; i++)
  {
    int randomIndex = indices[i];
    vfd->setCursor(randomIndex % DISPLAY_CHARS_X, randomIndex / DISPLAY_CHARS_X);
    vfd->write(useSpaces ? ' ' : (randomIndex + ' ' + 1));
    delay(25);
  } 
}

bool setCursorTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  setCursorTestFill(vfd, false);
    
  delay(4000);

  setCursorTestFill(vfd, true);
  
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool characterSetsTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  int x = 0;
  
  for (int i =' '; i<=255; i++)
  {
    x = (i-' ')  % DISPLAY_CHARS_X;
    
    vfd->setCursor(x, 0);
    vfd->setDefaultCharacterSet();
    vfd->write(i);

    vfd->setCursor(x, 1);
    vfd->setJapaneseCharacterSet();
    vfd->write(i);

    vfd->setCursor(x, 2);
    vfd->setCodePage850CharacterSet();
    vfd->write(i);
    
    delay(250);
  }

  vfd->setDefaultCharacterSet();

  for (x++;x < DISPLAY_CHARS_X; x++)
  {
    vfd->setCursor(x, 0);
    vfd->write(' ');
    
    vfd->setCursor(x, 1);
    vfd->write(' ');

    vfd->setCursor(x, 2);
    vfd->write(' ');
  }

  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

void verticalScrollPrintLoop(Ncr4X20Vfd *vfd, const char *labelString)
{
  for (int i=1; i<=12; i++)
  {
    vfd->print("Row ");
    vfd->print(i);
    delay(250);
    vfd->print(labelString);
  }
}

bool verticalScrollTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();
  
  clearAndHomeDisplay(vfd);
  
  for (int i=0; i<2; i++)
  {
    vfd->verticalScroll();
    verticalScrollPrintLoop(vfd, " Scrolling \r\n");
   
    vfd->noVerticalScroll();
    verticalScrollPrintLoop(vfd, " Wrapping  \r\n");
  }
  
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool startDiagnosticSequenceTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();

  clearAndHomeDisplay(vfd);
  
  vfd->startDiagnosticSequence();
  
  vfd->print("You should not see\r\nthis.");

  delay(15000);

  clearAndHomeDisplay(vfd);
  vfd->print("Ended diagnostic\r\nsequence.");
  
  vfd->endDiagnosticSequence();
  
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

void highBitCharacterTestPrintCharacters(Ncr4X20Vfd *vfd, int offset)
{
  for (int i='A'; i <='Z'; i++)
  {
    vfd->write(i+offset);
  }

  vfd->write(' ');

  for (int i='0'; i <='9'; i++)
  {
    vfd->write(i+offset);
  }
}

bool highBitCharacterTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();

  clearAndHomeDisplay(vfd);

  // Print characters with bit 7 not set.
  highBitCharacterTestPrintCharacters(vfd, 0);
  
  vfd->print("\r\n");
  
  // Print same characters with bit 7 set.
  highBitCharacterTestPrintCharacters(vfd, 128);
  
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

void loadTestPrintLastLine(Ncr4X20Vfd *vfd)
{
  for (int i='A'; i<='T'; i++)
  {
    vfd->write(i);
  }
}

bool loadTestSequenceCharactersTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();

  clearAndHomeDisplay(vfd);

  for (int character=' '; character<=255; character++)
  {
    for (int i=0; i<DISPLAY_CHARS_TOTAL; i++)
    {
      vfd->write(character);
    }
  }
 
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool loadTestWithClearsTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();

  clearAndHomeDisplay(vfd);

  for (int i=0; i<100; i++)
  {
    for (int i=0; i<DISPLAY_CHARS_TOTAL; i++)
    {
      vfd->write((i % 10) + '0');
      vfd->clear();      
    }
  }
 
  loadTestPrintLastLine(vfd); 
 
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

bool loadTestWithScrollingTestFunction(bool useBusyPin, bool useData7Pin)
{
  Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
  vfd->begin();

  clearAndHomeDisplay(vfd);
  
  vfd->verticalScroll();
  
  for (int i=0; i<100; i++)
  {
    for (int i=0; i<DISPLAY_CHARS_TOTAL; i++)
    {
      vfd->write((i % 10) + '0');
      vfd->write('\n');
    }
  }
  
  vfd->noVerticalScroll();
 
  loadTestPrintLastLine(vfd); 
 
  bool hadError = vfd->getError();
  delete vfd;
  return hadError;
}

#if USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY
bool powerCycleDisplay()
{
  bool hadError = false;
  
  digitalWrite(DISPLAY_POWER_SWITCH_PIN, HIGH); // Turn off display.
 
  delay(500); // Wait half a second.
  
  // Make sure all status lines are low.
  if (digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS1_PIN) != LOW ||
      digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS2_PIN) != LOW ||
      digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS3_PIN) != LOW ||
      digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS4_PIN) != LOW)
  {
    Serial.print("ERROR: Status lines still high, display did not power down?\n");
    
    hadError = true;
  }
  
  // Power display back on.
  digitalWrite(DISPLAY_POWER_SWITCH_PIN, LOW);

  return hadError;
}
#else // USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY
bool powerCycleDisplay()
{
  bool hadError = false;

  Serial.print("Power off display only (not Arduino).\n");
  
  unsigned long timeoutStart = millis();
  
  while (digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS1_PIN) != LOW ||
         digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS2_PIN) != LOW ||
         digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS3_PIN) != LOW ||
         digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS4_PIN) != LOW)
  {
    if ((millis() - timeoutStart) > 15000)
    {
      Serial.print("ERROR: Timed out waiting for display to be powered down.\n");
      hadError |= true;
      break;
    }
  }

  if (!hadError)
  {
    Serial.print("Power on display.\n");
    
    unsigned long timeoutStart = millis();

    while (digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS1_PIN) == LOW &&
           digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS2_PIN) == LOW &&
           digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS3_PIN) == LOW &&
           digitalRead(NCR_4X20_VFD_UNKNOWN_STATUS4_PIN) == LOW)
    {
      if ((millis() - timeoutStart) > 15000)
      {
        Serial.print("ERROR: Timed out waiting for display to be powered up.\n");
        hadError |= true;
        break;
      }
    }
  }

  if (!hadError)
  {
    Serial.print("Test continuing...\n");
  }
  
  return hadError;
}
#endif // USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY

bool hardResetTestFunction(bool useBusyPin, bool useData7Pin)
{
  bool hadError = false;
  
  hadError |= powerCycleDisplay();
  if (!hadError)
  {
    Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
    
    vfd->reInitStartupTimeouts(); // Make sure the library waits for the hardware to power up properly.
    
    vfd->begin();

    vfd->print("YOU SHOULD SEE THIS.");
  
    hadError |= vfd->getError();
    delete vfd;
  }

  return hadError;
}

bool forceTimeoutTestFunction(bool useBusyPin, bool useData7Pin)
{
  bool hadError = false;
  
  if (!useBusyPin)
  {
    Serial.print("This test is not applicable when not using the busy pin.\n");
  }
  else
  { 
    Ncr4X20Vfd *vfd = createVfd(useBusyPin, useData7Pin);
    vfd->begin();

    clearAndHomeDisplay(vfd); // This will clear all long timeouts.
 
    // Hard reset display.
    hadError |= powerCycleDisplay();
    if (!hadError)
    {
      delay(100); // Wait for display to stabalize. 
      
      // Display should now be in startup sequence and return busy for several seconds.
      vfd->print("YOU SHOULD NOT SEE\r\nTHIS."); // This print should fail. 
      if (vfd->getError())
      {
        // An error was correctly returned.
        if (vfd->getError())
        {
          // The error state did not cleared itself as it should have.
          Serial.print("ERROR: GetError() status did not clear on second read.\n");
          hadError |= true;
        }
        else
        {
          Serial.print("Timeouts are working correctly.\n");
        }
      }
      else
      {
        Serial.print("ERROR: Did not get expected timeout error.\n");
        hadError |= true;
      }
      
      delay(Ncr4X20Vfd::STARTUP_TIME_IN_MILLISECONDS);
      
      vfd->print("YOU SHOULD SEE THIS."); // This print should pass. 
    }
    
    hadError |= vfd->getError();
    delete vfd;
  }
  
  return hadError;
}
    
UnitTest unitTests[] =
{
  {
    basicTextTestTitle,
    basicTextTestInstructions,
    basicTextTestFunction
  },
  {
    clearTestTitle,
    clearTestInstructions,
    clearTestFunction
  },
  {
    homeTestTitle,
    homeTestInstructions,
    homeTestFunction
  },
  {
    noDisplayTestTitle,
    noDisplayTestInstructions,
    noDisplayTestFunction
  },
  {
    blinkCursorTestTitle,
    blinkCursorTestInstructions,
    blinkCursorTestFunction
  },
  {
    setCursorTestTitle,
    setCursorTestInstructions,
    setCursorTestFunction,
  },
  {
    characterSetsTestTitle,
    characterSetsTestInstructions,
    characterSetsTestFunction
  },
  {
    verticalScrollTestTitle,
    verticalScrollTestInstructions,
    verticalScrollTestFunction
  },
  {
    startDiagnosticSequenceTestTitle,
    startDiagnosticSequenceTestInstructions,
    startDiagnosticSequenceTestFunction
  },
  {
    highBitCharacterTestTitle,
    highBitCharacterTestInstructions,
    highBitCharacterTestFunction
  },
  {
    loadTestSequenceCharactersTestTitle,
    loadTestSequenceCharactersTestInstructions,
    loadTestSequenceCharactersTestFunction
  },
  {
    loadTestWithClearsTestTitle,
    loadTestWithClearsTestInstructions,
    loadTestWithClearsTestFunction
  },
  {
    loadTestWithScrollingTestTitle,
    loadTestWithScrollingTestInstructions,
    loadTestWithScrollingTestFunction
  },
  {
    hardResetTestTitle,
    hardResetTestInstructions,
    hardResetTestFunction
  },
  {
    forceTimeoutTestTitle,
    forceTimeoutTestInstructions,
    forceTimeoutTestFunction
  },
  {
     NULL,
     NULL,
     NULL  
  }
};
    
void setup()
{
  // initialize serial communications at 9600 bps.
  Serial.begin(9600); 

  // Drive the unneeded control lines high. The display's parallel 
  // interface will not function correctly if they are low.
  digitalWrite(NCR_4X20_VFD_RESET_PIN, HIGH);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_CONTROL1_PIN, HIGH);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_CONTROL2_PIN, HIGH);
  
  pinMode(NCR_4X20_VFD_RESET_PIN, OUTPUT);
  pinMode(NCR_4X20_VFD_UNKNOWN_CONTROL1_PIN, OUTPUT);
  pinMode(NCR_4X20_VFD_UNKNOWN_CONTROL2_PIN, OUTPUT);   

  // Set uknown control pins to inputs with pullup resistor disabled.
  // (Overkill, this is the default state).
  pinMode(NCR_4X20_VFD_UNKNOWN_STATUS1_PIN, INPUT);   
  pinMode(NCR_4X20_VFD_UNKNOWN_STATUS2_PIN, INPUT);   
  pinMode(NCR_4X20_VFD_UNKNOWN_STATUS3_PIN, INPUT);   
  pinMode(NCR_4X20_VFD_UNKNOWN_STATUS4_PIN, INPUT);   
  
  digitalWrite(NCR_4X20_VFD_UNKNOWN_STATUS1_PIN, LOW);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_STATUS2_PIN, LOW);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_STATUS3_PIN, LOW);
  digitalWrite(NCR_4X20_VFD_UNKNOWN_STATUS4_PIN, LOW);

#if USE_TRANSISTOR_TO_POWER_CYCLE_DISPLAY
  // Set power switch pin low. If the switch is connected,
  // it is assumed to be PNP/p-channel and will therefore
  // be switched on by a low signal.
  digitalWrite(DISPLAY_POWER_SWITCH_PIN, LOW);
  pinMode(DISPLAY_POWER_SWITCH_PIN, OUTPUT);  
#endif

  // Wait for a response from serial before we start.
  for (;;)
  {
    if (Serial.available()) 
    {
      byte character = Serial.read();
      if (character == 'c' || character == 'C')
      {
        break;
      }
    }

    Serial.print("Enter 'c' to continue.\n");
      
    delay(1000);
  }
}

void loop()
{  
  static bool useBusyPin = true;
  static bool useData7Pin = true;

  Serial.print("\n"
               "Select test or option:\n"
               "\n"
               "b Toggle use of busy pin ");
  Serial.print(useBusyPin ? "(currently used)\n" : "(currently not used)\n");
  Serial.print("d Toggle use of data 7 pin ");
  Serial.print(useData7Pin ? "(currently used)\n" : "(currently not used)\n");  
  Serial.print("\n");
    
  int numTests = 0;
  for (int i=0; unitTests[i]._title != NULL; i++, numTests++)
  {
    Serial.print(i+1);
    Serial.print(" ");
    serialPrintFromProgmem(unitTests[i]._title);
    Serial.print("\n");
  }
  
  char inputBuffer[4];
  
  getString(inputBuffer, sizeof(inputBuffer)/sizeof(char));
  
  if ((inputBuffer[0] == 'b' || inputBuffer[0] == 'B') && inputBuffer[1]==0)
  {
    useBusyPin = useBusyPin ? false : true;
  }
  else if ((inputBuffer[0] == 'd' || inputBuffer[0] == 'D') && inputBuffer[1]==0)
  {
    useData7Pin = useData7Pin ? false : true;
  }
  else
  {
    int testNumber = atoi(inputBuffer) - 1;
    if (testNumber>=0 && testNumber<numTests)
    { 
      // A valid test was selected.
      Serial.print("\n");
      Serial.print("Test ");
      Serial.print(testNumber+1);
      Serial.print(" selected\n\n");
      serialPrintFromProgmem(unitTests[testNumber]._instructions);
      Serial.print("\n"
                   "Enter 'c' to continue.\n");
    
      // Wait for 'c' to be sent.
      for (;;)
      {
        if (Serial.available()) 
        {
          byte character = Serial.read();
          if (character == 'c' || character == 'C')
          {
            break;
          }
        }
      }

      Serial.print("\n");

      // Run the test.
      bool hadError = unitTests[testNumber]._testFunction(useBusyPin, useData7Pin);
      
      Serial.print("\n");
      Serial.print(hadError ? "ERROR ERROR ERROR The test reported an error!!!\n" : "Test completed.\n");
    }    
  }
}

// This is to force the sketch to use more than 16k of program memory
// so it can't be run on a lesser microcontroller having less than 2k 
// of SRAM.
prog_char dummyData [] PROGMEM = 
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111"
  "1111111111111111111111111111111111111111111111111111111111111111";

